from datetime import datetime
from secrets import token_bytes
from AsymmetricEncryptions.Protocols import KDF
import json
from .KDC import KDC
from base64 import b64decode

class KerberosClient:
    from AsymmetricEncryptions.General.Feistel_sha256 import FeistelSha256

    def __init__(self, c_id: str, passwd: bytes, dec_func: callable = FeistelSha256.get_feistel().decrypt, symmetric_enc_func: callable = FeistelSha256.get_feistel().encrypt):
        self.id = c_id
        self.passwd = passwd
        self.key = KDF.derive_key(passwd)
        self.dec = KDC.dec_with_hmac(dec_func)
        self.enc = KDC.add_hmac(symmetric_enc_func)



    def approach_AS(self, service_id: str) -> str and int:
        from AsymmetricEncryptions.General import BytesAndInts
        nonce = BytesAndInts.byte2Int(token_bytes(32))
        d = {"service_id": service_id, "client_id": self.id, "nonce": nonce}
        return json.dumps(d), nonce

    def approach_TGS(self, msgA: bytes, TGT: bytes) -> (bytes, bytes) and bytes:
        # get Kc-tgs
        msgA_dec = self.dec(msgA, self.key)
        msgA_dict = json.loads(msgA_dec)
        Kc_tgs: bytes = b64decode(msgA_dict["Kc_TGS"])
        msgD_dict = {"client_id": self.id, "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
        msgD_json = json.dumps(msgD_dict)
        msgD = self.enc(msgD_json, Kc_tgs)
        return (TGT, msgD), Kc_tgs

    def approach_service(self, ticket: bytes, msgF: bytes, Kc_tgs):
        Kc_s = b64decode(self.dec(msgF, Kc_tgs))
        msgG_dict = {"client_id": self.id, "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
        msgG_json = json.dumps(msgG_dict)
        msgG = self.enc(msgG_json, Kc_s)
        return (ticket, msgG), Kc_s





__all__ = ["EllipticCurveNISTP256", "ECPoint", "ECDH", "ECKey", "ECSchnorr", "ECIES", "ECCurve", "ECMQV", "ECDSA"]
from .ECPoint import ECPoint
from .ECDH import ECDH
from .ECKey import ECKey
from .ECSchnorr import ECSchnorr
from .ECIES import ECIES
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
from .ECCurve import ECCurve
from .ECMQV import ECMQV
from .ECDSA import ECDSA



from __future__ import annotations
from abc import ABC, abstractmethod

class ISignAndVerify(ABC):

    @staticmethod
    @abstractmethod
    def generate_key_pair(nBit: int):
        raise NotImplementedError("Function should be implemented inside of class")

    @abstractmethod
    def sign(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")

    @abstractmethod
    def verify(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")







__all__ = ["IKey", "IEncryptAndDecrypt", "ISignAndVerify", "IExport"]
from .IKey import IKey
from .IEncryptAndDecrypt import IEncryptAndDecrypt
from .ISignAndVerify import ISignAndVerify
from .IExport import IExport


from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSAKey

class TestRSAKey(TestCase):
    def test_new(self):
        priv = RSAKey.new(256)
        pub: RSAKey = priv.public
        self.assertEqual((pub.e * priv.d) % priv.tot_n, 1)

    def test_new_values(self):
        self.assertRaises(ValueError, RSAKey.new, -1)
        self.assertRaises(ValueError, RSAKey.new, 4209)

    def test_export_and_load(self):
        priv: RSAKey = RSAKey.new(256)
        priv.export("TestFiles\\RSAKeyTestFileUnprotected.key", b"")
        loaded_priv = RSAKey.load("TestFiles\\RSAKeyTestFileUnprotected.key", b"")
        self.assertEqual(priv, loaded_priv)
        priv: RSAKey = RSAKey.new(256)
        priv.export("TestFiles\\RSAKeyTestFileProtected.key", b"Pass")
        loaded_priv = RSAKey.load("TestFiles\\RSAKeyTestFileProtected.key", b"Pass")
        self.assertEqual(priv, loaded_priv)






from __future__ import annotations
from AsymmetricEncryptions.Protocols.OAEP import OAEP
from unittest import TestCase

class TestOAEP(TestCase):

    def test_OAEP(self):
        msg = b"OAEP"
        padded = OAEP.oaep_pad(msg)
        unpadded = OAEP.oaep_unpad(padded)
        self.assertEqual(msg, unpadded)
        self.assertNotEqual(msg, padded)

    def test_ErrOAEP(self):
        msg = b"OAEP" * 8 + b"a"
        self.assertRaises(ValueError, OAEP.oaep_pad, msg)














from __future__ import annotations
from unittest import TestCase
from AsymmetricEncryptions.Protocols.SSS import SSS
from secrets import token_bytes, SystemRandom


class TestSSS(TestCase):

    def test_init(self):
        self.assertRaises(ValueError, SSS, -1)
        self.assertRaises(ValueError, SSS, 4)

    def test_make_random_shares_and_recover(self):
        secret = token_bytes(16)
        mini = SystemRandom().randint(3, 16)
        maxi = SystemRandom().randint(16, 20)
        sss = SSS(mini=mini, maxi=maxi)
        shares = sss.make_random_shares(secret, mini, maxi)
        recover = sss.recover_secret(shares[:mini])
        self.assertEqual(recover, secret)
        self.assertRaises(ValueError, sss.recover_secret, shares[:(mini - 1)])
        self.assertRaises(ValueError, sss.recover_secret, [shares[0], shares[0]])
        self.assertRaises(ValueError, sss.make_random_shares, 4.1, -1, -2)

    def test_value(self):
        self.assertRaises(ValueError, SSS(3).make_random_shares, b"test"*5, 4, 5)

# secret = b"test"
# sss = SSS()
# shares = sss.make_random_shares(secret, 3, 5)
# recover = sss.recover_secret(shares[:3])
# print(recover)


from unittest import TestCase
from secrets import randbelow
from AsymmetricEncryptions.MPC.DiningCryptographers import Cryptographer

class TestDC(TestCase):

    def test_dining_protocol_not_paid(self):
        n = randbelow(20)
        while n < 3:
            n = randbelow(20)
        cryptographers = [Cryptographer(n, i, name=f"{i}") for i in range(n)]
        for cryptographer in cryptographers:
            cryptographer.other_cryptographers = cryptographers.copy()
            cryptographer.establish_bits()
        self.assertEqual(cryptographers[randbelow(n)].dine(), 0)

    def test_dining_protocol_paid(self):
        n = randbelow(20)
        while n < 3:
            n = randbelow(20)
        someone_paid = False
        cryptographers = []
        for i in range(n):
            pay = False
            if not someone_paid:
                pay = bool(randbelow(2))
                if pay: someone_paid = True
            cryptographers.append(Cryptographer(n, i, paid=pay, name=f"{i}"))
        for cryptographer in cryptographers:
            cryptographer.other_cryptographers = cryptographers.copy()
            cryptographer.establish_bits()
        self.assertEqual(cryptographers[randbelow(n)].dine(), int(someone_paid))



from secrets import randbelow

class MPCGeneral:

    @staticmethod
    def make_shares(p: int, n: int, m: int = 0) -> list[int]:
        l = []
        c = p
        if m > 0:
            for _ in range(n - 1):
                r = randbelow(m)
                c = (c - r) % m
                l.append(r)
            l.append(c)
        else:
            for _ in range(n - 1):
                r = randbelow(p)
                c = (c - r)
                l.append(r)
            l.append(c)
        return l.copy()

    @staticmethod
    def combine_shares(shares: list[int], m: int = 0) -> int:
        s = 0
        if s != 0:
            for share in shares:
                s = (s + share) % m
        else:
            for share in shares:
                s = (s + share)
        return s



import math
def Log2(x):
    if x == 0:
        return False

    return (math.log10(x) /
            math.log10(2))

# Function to check
# if x is power of 2


def isPowerOfTwo(n):
    return (math.ceil(Log2(n)) ==
            math.floor(Log2(n)))


from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ECC import ECKey, EllipticCurveNISTP256
from AsymmetricEncryptions.Protocols.AOSRingSignatures import AOS
from secrets import randbelow, token_bytes, SystemRandom

class TestAOS(TestCase):

    def test_ring(self):
        m = token_bytes(16)
        n = randbelow(20)
        while n < 2:
            n = randbelow(20)
        signer = ECKey.new(curve := EllipticCurveNISTP256.get_curve())
        keys = [ECKey.new(curve).get_public_key() for _ in range(n)]
        rs = AOS(keys, signer)
        sigma, rk = rs.sign(m)
        v = AOS.verify(rk, m, sigma)
        self.assertEqual(v, True)
        v = AOS.verify(rk, m+b" Ring", sigma)
        self.assertEqual(v, False)
        nrk = rk.copy()
        SystemRandom().shuffle(nrk)
        v = AOS.verify(nrk, m, sigma)
        self.assertEqual(v, False)
        v = AOS.verify(rk, m, (sigma[0] + 1, sigma[1]))
        self.assertEqual(v, False)
        v = AOS.verify(rk, m, (sigma[0], sigma[1] + [2]))
        self.assertEqual(v, False)
        sigma[1][0] = 1
        v = AOS.verify(rk, m, (sigma[0], sigma[1]))
        self.assertEqual(v, False)


import secrets
from unittest import TestCase
from unittest.mock import patch, Mock
from AsymmetricEncryptions.General.Exportation import Exportation

class TestExportation(TestCase):

    @patch("AsymmetricEncryptions.General.Exportation.sha_wrapper")
    def test_export(self, mock_randomness):
        mock_randomness.return_value = b" "
        d = {"test": 123456789, "e": 69}
        self.assertEqual(b'3MCFp+lsJbLhsVPcCxAR1bzqBPbaabQWbyN8+s3qyatLgrCv020yik/0odPzGqNiee5Vj7+SdQT7Dcow8qHqA9LuGHzBSqUryeLfG/PCCdIGv/ZIMvrRdZac3XY6QU3I4DT0TEwo1mxi6FHRXVIFphPcFQ6B58mJg3OVtyNEHS8g', Exportation.export("testdict.txt", d, b"", isTesting=True))



    def test_export_and_load(self):
        d = {"test": 123456789, "e": 69}
        e = Exportation.export("testdict.txt", d, b"", isTesting=True)
        nd = Exportation.load("testdict.txt", b"", isTesting=True, dataIfTesting=e)
        self.assertEqual(d, nd)



from .MPC_General import MPCGeneral
class MPCAddition:

    def __init__(self, p: int, n: int, m: int = 0):
        self.p = p
        self.m = m
        self.n = n
        self.shares = MPCGeneral.make_shares(p, n, m)

    def get_shares(self) -> list[int]:
        return self.shares.copy()



    def s(self, shared: list[int]) -> int:
        return MPCGeneral.combine_shares(shared, self.m)



import json
import secrets
from AsymmetricEncryptions.Protocols import KDF
from AsymmetricEncryptions.Exceptions import ReplayAttack
from base64 import b64encode, b64decode
from datetime import datetime
from hashlib import sha256

class TGS:
    # Ticket Granting Server
    def __init__(self, services: dict[str, bytes], symmetric_enc_func, symmetric_dec_func, KTGS: bytes):
        self.services = services
        self.dec = symmetric_dec_func
        self.enc = symmetric_enc_func
        self.KTGS = KTGS
        self.cache = set()

    def grant_ticket(self, TGT: bytes, msgD: bytes):
        hash_ticket = sha256(TGT + msgD).hexdigest()
        if hash_ticket in self.cache: raise ReplayAttack()
        self.cache.add(hash_ticket)
        d_TGT = self.dec(TGT, self.KTGS)
        try:
            TGT = json.loads(d_TGT)
        except Exception:
            return None
        Kc_TGS = b64decode(TGT["Kc_TGS"].encode())
        lifetime = datetime.strptime(TGT["lifetime"], "%Y-%m-%d %H:%M:%S").replace(microsecond=0)
        assert datetime.now() < lifetime, "Life time has passed"
        verify_msgD = self.dec(msgD, Kc_TGS)
        msgD_dict = json.loads(verify_msgD)
        assert msgD_dict["client_id"] == TGT["client_id"]
        Kc_s = b64encode(KDF.derive_key(secrets.token_bytes(32))).decode()
        msgE = {"Kc_s": Kc_s, "client_id": TGT["client_id"], "lifetime": TGT["lifetime"], "nonce": TGT["nonce"]}
        msgE = json.dumps(msgE)
        Ks = KDF.derive_key(self.services[TGT["service_id"]])
        msgE = self.enc(msgE, Ks)
        msgF = self.enc(Kc_s, Kc_TGS)
        return msgE, msgF


from secrets import randbelow

class Cryptographer:

    def __init__(self, n: int, position=0, other_cryptographers=None, paid=False, name="A"):
        self.n = n
        self._paid = paid
        self.position = position
        if other_cryptographers is None: other_cryptographers = []
        self.other_cryptographers = other_cryptographers
        self.bits = [None] * n
        self.name = name



    def establish_bits(self):
        for i, cryptographer in enumerate(self.other_cryptographers):
            if self.other_cryptographers[i] == self: continue
            b = randbelow(2)
            self.bits[i] = b
            cryptographer.bits[self.position] = b

    def __str__(self):
        return f"{self.name}"


    def announce(self):
        self.bits[self.position] = int(self._paid)
        x = 0
        for i in self.bits:
            x = x ^ i
        return x

    def dine(self):
        bits = []
        for cryptographer in self.other_cryptographers:
            bits.append(cryptographer.announce())
        x = 0
        for i in bits:
            x = x ^ i
        return x



from __future__ import annotations
from AsymmetricEncryptions.Protocols.Padding import PKCS7
from unittest import TestCase

class TestPKCS7(TestCase):

    def test_pad(self):
        msg = b"PKCS7"
        block_size = 32
        padded = PKCS7(block_size).pad(msg)
        self.assertEqual(b'PKCS7\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b', padded)

    def test_unpad(self):
        unpadded = PKCS7(32).unpad(b'PKCS7\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b\x1b')
        self.assertEqual(unpadded, b"PKCS7")


from AsymmetricEncryptions.PublicPrivateKey.ECC import ECMQV, ECKey, EllipticCurveNISTP256

from unittest import TestCase

class TestMQV(TestCase):

    def test_MQV(self):
        a = ECKey.new(EllipticCurveNISTP256.get_curve())
        b = ECKey.new(EllipticCurveNISTP256.get_curve())

        x = ECMQV.Stage1n2(EllipticCurveNISTP256.get_curve())
        y = ECMQV.Stage1n2(EllipticCurveNISTP256.get_curve())

        Sa = ECMQV.Stage3n4(a, x)
        Sb = ECMQV.Stage3n4(b, y)

        keyA = ECMQV.Stage5(b.get_public_key(), Sa, y.get_public_key())
        keyB = ECMQV.Stage5(a.get_public_key(), Sb, x.get_public_key())
        self.assertEqual(keyA, keyB)


from __future__ import annotations
from .ElGamalKey import ElGamalKey
from AsymmetricEncryptions.General import BytesAndInts
from AsymmetricEncryptions.Exceptions import Exceptions
from AsymmetricEncryptions.Interfaces import ISignAndVerify, IEncryptAndDecrypt
import secrets
from math import gcd
import hashlib

class ElGamal(ISignAndVerify, IEncryptAndDecrypt):
    """ElGamal encryption"""
    def __init__(self, key: ElGamalKey) -> None:
        """
        :param key: private / public key
        """
        self.key = key

    @staticmethod
    def generate_key_pair(nBit) -> ElGamalKey and ElGamalKey:
        """
        ElGamalKey.new returns a pair of priv and pub
        :param nBit: How big is p.
        :return: private_key, public_key
        """
        Priv: ElGamalKey = ElGamalKey.new(nBit)
        Pub: ElGamalKey = Priv.public
        return Priv, Pub

    def encrypt(self, msg: bytes, assertion=True) -> tuple[bytes, bytes]:
        """
        Encrypts a message
        :param msg: The plaintext
        :param assertion: If assertion is true then it will assert if the message is smaller than allowed. (e.g. m < p)
        :return: Ciphertext
        """

        m: bytes = BytesAndInts.byte2Int(msg)
        if assertion: assert m < self.key.p
        b: int = secrets.randbelow(self.key.p)
        c1: int = pow(self.key.g, b, self.key.p)
        c2: int = (m * pow(self.key.y, b, self.key.p)) % self.key.p
        return BytesAndInts.int2Byte(c1), BytesAndInts.int2Byte(c2)

    def decrypt(self, cipher: tuple[bytes, bytes]) -> bytes:
        """
        Decrypts (Only possible if key is private)
        :param cipher: the ciphertext
        :return: message
        """
        assert self.key.x
        c1: int = BytesAndInts.byte2Int(cipher[0])
        c2: int = BytesAndInts.byte2Int(cipher[1])
        a: int = pow(c1, self.key.x, self.key.p)
        m = (c2 * pow(a, self.key.p - 2, self.key.p)) % self.key.p
        return BytesAndInts.int2Byte(m)


    def sign(self, msg: bytes, assertion=True) -> tuple[bytes, bytes, bytes]:
        """
        Signs a message (turns to hash)
        @param msg: The plaintext
        @param assertion: if true then will check if m is allowed
        @return: signature
        """
        msg = hashlib.sha256(msg).digest()
        m: bytes = BytesAndInts.byte2Int(msg)
        if assertion: assert m < self.key.p
        k: int = secrets.randbelow(self.key.p)
        while gcd(k, self.key.p - 1) != 1: k = secrets.SystemRandom().randint(2, self.key.p - 2)
        s1: int = pow(self.key.g, k, self.key.p)
        phi_n: int = self.key.p - 1
        inv: int = pow(k, -1, phi_n)
        s2: int = (inv * (m - s1 * self.key.x)) % phi_n
        if s2 == 0: self.sign(msg, assertion)
        return BytesAndInts.int2Byte(s1), BytesAndInts.int2Byte(s2),  BytesAndInts.int2Byte(m)

    def verify(self, signature: tuple[bytes, bytes, bytes], *, og_message: bytes = None) -> bool:
        """
        Verifies a signature
        @param og_message: The original message
        @param signature: The signature
        @return: bool
        """
        if og_message:
            assert signature[2] == hashlib.sha256(og_message).digest()
        s1: int = BytesAndInts.byte2Int(signature[0])
        s2: int = BytesAndInts.byte2Int(signature[1])
        m: int = BytesAndInts.byte2Int(signature[2])
        V = pow(self.key.y, s1, self.key.p) * pow(s1, s2, self.key.p)
        V = V % self.key.p
        W = pow(self.key.g, m, self.key.p)
        if not V == W:
            raise Exceptions.MACError("Signature doesn't match!")



__all__ = ["AuthServer", "TicketGrantingServer", "KRBS_Client", "KDC", "KRBS_Service"]
from .AuthServer import AS
from .TicketGrantingServer import TGS
from .KDC import KDC
from .KRBS_Client import KerberosClient
from .KRBS_Service import KerberosService



from AsymmetricEncryptions.Protocols.RSADH import DiffieHellman
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSA
from unittest import TestCase

class TestRSADH(TestCase):

    def test_DiffieHellman(self):
        Apriv, Apub = RSA.generate_key_pair(256)
        Bpriv, Bpub = RSA.generate_key_pair(256)
        DH = DiffieHellman.new(Apriv, 256)
        A = DH.Stage1()
        gp = DH.get_gp()
        B, Bob_shared = DiffieHellman.Stage2(gp, Bpriv, A)
        Alice_shared = DH.Stage3(B)
        self.assertEqual(Alice_shared,Bob_shared)





from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.DLIES import DLIES
from secrets import token_bytes

class TestDLIES(TestCase):

    def test_generate_key_pair(self):
        priv, pub = DLIES.generate_key_pair(256)
        self.assertEqual(pub.y, pow(pub.g, priv.x, pub.n))

    def test_generate_key_pair_type(self):
        self.assertRaises(TypeError, DLIES.generate_key_pair, "test")
        self.assertRaises(TypeError, DLIES.generate_key_pair, b"test")

    def test_generate_key_pair_values(self):
        self.assertRaises(ValueError, DLIES.generate_key_pair, -2)
        self.assertRaises(ValueError, DLIES.generate_key_pair, 5)

    def test_encrypt(self):
        priv, pub = DLIES.generate_key_pair(256)
        c = DLIES.encrypt(pub, t := token_bytes(1))
        self.assertEqual(t, DLIES.decrypt(priv, c))

    def test_encrypt_type(self):
        priv, pub = DLIES.generate_key_pair(256)
        self.assertRaises(TypeError, DLIES.encrypt, pub, "test")
        self.assertRaises(TypeError, DLIES.encrypt, "test", "test")

    def test_decrypt(self):
        priv, pub = DLIES.generate_key_pair(256)
        c = DLIES.encrypt(pub, b"test")
        self.assertEqual(b"test", DLIES.decrypt(priv, c))

    def test_signature_and_verification(self):
        priv, pub = DLIES.generate_key_pair(1024)
        signer = DLIES(priv, pub)
        m = b"test"
        sig = signer.sign(m)
        v = DLIES.verify(pub, sig, m)
        self.assertEqual(v, True)
        v = DLIES.verify(pub, sig, m+b"no")
        self.assertEqual(v, False)



from __future__ import annotations
from abc import ABC, abstractmethod

class IExport(ABC):
    @abstractmethod
    def export(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")

    @staticmethod
    @abstractmethod
    def load(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")



from __future__ import annotations
# MQV (Menezes–Qu–Vanstone) is an authenticated protocol for key agreement based on the Diffie–Hellman scheme. Like other authenticated Diffie–Hellman schemes, MQV provides protection against an active attacker. The protocol can be modified to work in an arbitrary finite group, and, in particular, elliptic curve groups, where it is known as elliptic curve MQV (ECMQV).
from .ECKey import ECKey
from .ECPoint import ECPoint
from .ECCurve import ECCurve
from math import floor, ceil, log
class ECMQV:


    @staticmethod
    def bar(point: ECPoint) -> int:
        L = ceil((floor(log(point.curve.n, 2)) + 1) / 2)
        TwoPowL = pow(2, L)
        return (point.x % TwoPowL) + TwoPowL

    @staticmethod
    def Stage1n2(curve: ECCurve) -> ECKey:
        return ECKey.new(curve)

    @staticmethod
    def Stage3n4(priv_key: ECKey, xy: ECKey) -> int:
        return (xy.private_key + (ECMQV.bar(xy.public_key) * priv_key.private_key) % priv_key.public_key.curve.n) % priv_key.public_key.curve.n

    @staticmethod
    def Stage5(other_pub_key: ECKey, S: int, xy: ECKey) -> ECPoint:
        return (xy.public_key + (other_pub_key.public_key * ECMQV.bar(xy.public_key))) * other_pub_key.public_key.curve.h * S



# Key Distribution Center
from datetime import datetime

from AsymmetricEncryptions.General.Feistel_sha256 import FeistelSha256
from AsymmetricEncryptions.Protocols import KDF
from AsymmetricEncryptions.Protocols.Kerberos import AS, TGS
from hashlib import sha256
import hmac
from build.lib.AsymmetricEncryptions import MACError
from AsymmetricEncryptions.General.XOR import XOR


class KDC:

    def __init__(self, clients: dict[str, bytes], services: dict[str, bytes], master_passwd: bytes, symmetric_enc_func=FeistelSha256.get_feistel().encrypt, symmetric_dec_func=FeistelSha256.get_feistel().decrypt, ticket_life_time=datetime(1, 1, 1, 1, minute=1)):
        self.clients = clients
        self.services = services
        self.master_passwd = KDF.derive_key(master_passwd)
        self.enc = self.add_hmac(symmetric_enc_func)
        self.dec = self.dec_with_hmac(symmetric_dec_func)
        self.KTGS = KDF.derive_key(self.master_passwd)
        self.ticket_life_time = ticket_life_time
        self.auth_server = AS(clients, self.enc, self.KTGS, ticket_life_time)
        self.ticket_granting_server = TGS(services, self.enc, self.dec, self.KTGS)

    @staticmethod
    def add_hmac(enc) -> callable:
        def wrapper(m, k):
            if isinstance(m, str): m = m.encode()
            c = enc(m, k)
            mac = hmac.new(k, c, sha256).digest()
            return mac + c
        return wrapper

    @staticmethod
    def dec_with_hmac(dec):
        def wrapper(c, k):
            if isinstance(c, str): c = c.encode()
            ciphertxt = c[32:]
            mac = c[:32]
            new_mac = hmac.new(k, ciphertxt, sha256).digest()
            if not hmac.compare_digest(mac, new_mac): raise MACError("")
            m = dec(ciphertxt, k)
            return m
        return wrapper


    def AS_response(self, approach: str):
        return self.auth_server.reply(approach)

    def TGS_response(self, TGT: bytes, msgD: bytes):
        return self.ticket_granting_server.grant_ticket(TGT, msgD)










from __future__ import annotations
from abc import ABC, abstractmethod
from .IExport import IExport
class IEncryptAndDecrypt(ABC):

    @staticmethod
    @abstractmethod
    def generate_key_pair(nBit: int):
        raise NotImplementedError("Function should be implemented inside of class")

    @abstractmethod
    def encrypt(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")

    @abstractmethod
    def decrypt(self, *args, **kwargs):
        raise NotImplementedError("Function should be implemented inside of class")







from AsymmetricEncryptions.Protocols.YAK import YAK
from AsymmetricEncryptions.PublicPrivateKey import DLIESKey
from unittest import TestCase

class TestYAK(TestCase):

    def test_YAK(self):
        yakAlice = YAK.new(256, False)
        gq = yakAlice.get_gq()
        yakBob = YAK(*gq)
        g, q = gq
        AlicePriv = DLIESKey.build(g, q)
        AlicePub = AlicePriv.public
        BobPriv = DLIESKey.build(g, q)
        BobPub = BobPriv.public
        x, AliceSend = yakAlice.stage_1()
        y, BobSend = yakBob.stage_1()
        sharedAlice = yakAlice.stage_2(x, AlicePriv, BobSend, BobPub)
        sharedBob = yakBob.stage_2(y, BobPriv, AliceSend, AlicePub)
        self.assertEqual(sharedAlice, sharedBob)


import secrets
import hashlib
import functools
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSAKey

class RingSignatures:
    """RSA implementation. Edited from wiki : https://en.wikipedia.org/wiki/Ring_signature"""

    def __init__(self, k: list[RSAKey], L: int = None) -> None:
        self.k = k
        if L is None:
            L = k[0].size
        self.l = L
        self.n = len(k)
        self.q = 1 << (L - 1)

    def sign_message(self, m: bytes, z: int) -> list[int] and list[RSAKey]:
        self._permut(m)
        s = [None] * self.n
        u = secrets.SystemRandom().randint(0, self.q)
        c = v = self._E(u)
        first_range = list(range(z + 1, self.n))
        second_range = list(range(z))
        whole_range = first_range + second_range

        for i in whole_range:
            s[i] = secrets.SystemRandom().randint(0, self.q)
            e = self._g(s[i], self.k[i].e, self.k[i].n)
            v = self._E(v ^ e)
            if (i + 1) % self.n == 0:
                c = v

        s[z] = self._g(v ^ u, self.k[z].d, self.k[z].n)
        rk = self.k.copy()
        rk[z] = rk[z].get_pub()
        return [c] + s, rk

    def verify_message(self, m: bytes, X: list[int]) -> bool:
        self._permut(m)

        def _f(i):
            return self._g(X[i + 1], self.k[i].e, self.k[i].n)

        y = map(_f, range(len(X) - 1))
        y = list(y)

        def _g(x, i):
            return self._E(x ^ y[i])

        r = functools.reduce(_g, range(self.n), X[0])
        return r == X[0]

    def _permut(self, msg: bytes):
        self.p = int(hashlib.sha256(msg).hexdigest(), 16)

    def _E(self, x):
        msg = f"{x}{self.p}".encode("utf-8")
        return int(hashlib.sha256(msg).hexdigest(), 16)

    def _g(self, x, e, n):
        q, r = divmod(x, n)
        if ((q + 1) * n) <= ((1 << self.l) - 1):
            result = q * n + pow(r, e, n)
        else:
            result = x
        return result


# An implementation of AOS ring signatures
from AsymmetricEncryptions.PublicPrivateKey.ECC import ECKey, ECPoint, ECCurve
from AsymmetricEncryptions.General.BytesAndInts import BytesAndInts
from hashlib import sha256
from secrets import randbelow, SystemRandom
from warnings import warn

class AOS:

    def __init__(self, keys: list[ECKey], signer: ECKey):
        self.keys = keys
        self.signer_key = signer

    def sign(self, m: bytes):
        curve = self.signer_key.curve
        keys = self.keys.copy()
        SystemRandom().shuffle(keys)
        j = randbelow(len(keys) + 1)
        keys.insert(j, self.signer_key)
        n = len(keys)
        s = [None] * n

        def H(M:bytes):
            return BytesAndInts.byte2Int(sha256(M + m).digest())

        alpha = randbelow(curve.n-1)
        g = curve.g()
        Q = g * alpha
        e = H(str(Q).encode())
        e0 = b""
        for i in range(j+1, n+j):
            indexI = i % n
            s[indexI] = randbelow(curve.n-1)
            e = H(str((g*s[indexI]) + (keys[indexI].public_key * e)).encode())
            if (indexI+1) % n == 0:
                e0 = e
        s[j] = (alpha - (e*keys[j].private_key)) % curve.n
        sigma = (e0, s)
        rk = keys.copy()
        rk[j] = rk[j].get_public_key()
        return sigma, rk.copy()

    @staticmethod
    def verify(rk: list[ECKey], m: bytes, sigma: tuple[int, list[int]]):
        s = sigma[1]
        e = e_0 = sigma[0]
        n = len(rk)
        if n != len(s):
            warn("Might be missing a key")
            return False
        curve = rk[0].curve

        def H(M: bytes) -> int:
            return BytesAndInts.byte2Int(sha256(M + m).digest())

        for indexI in range(n):
            e = H(str((curve.g()*s[indexI]) + (rk[indexI].public_key * e)).encode())

        return e == e_0



from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ECC import EllipticCurveNISTP256, ECKey, ECPoint, ECDSA
from secrets import randbelow
class TestECDSA(TestCase):
    def test_sign_and_verify(self):
        # Do this
        priv = ECKey.new(curve := EllipticCurveNISTP256.get_curve())
        pub = priv.get_public_key()
        m = b"Super cool message"
        signer = ECDSA(priv, pub)
        sig = signer.sign(m)
        v = ECDSA.verify(m, pub, sig)
        self.assertEqual(v, True)
        priv = ECKey.new(curve)
        pub = priv.get_public_key()
        m = b"Super cool message"
        signer = ECDSA(priv, pub)
        sig = signer.sign(m)
        v = ECDSA.verify(m+b"Not so cool", pub, sig)
        self.assertEqual(v, False)
        priv = ECKey.new(curve)
        pub = priv.get_public_key()
        m = b"Super cool message"
        signer = ECDSA(priv, pub)
        sig = signer.sign(m)
        sig = (sig[0], -1)
        v = ECDSA.verify(m, pub, sig)
        self.assertEqual(v, False)
        sig = (-1, sig[1])
        v = ECDSA.verify(m, pub, sig)
        self.assertEqual(v, False)

    def test_find_priv(self):
        # Never, ever, do this. this is to show that a misuse can cause a private key leakage. k needs to be random. You could (And maybe should) pass a hash of H(m || private_key) through a PRNG and put k as that.
        priv = ECKey.new(curve := EllipticCurveNISTP256.get_curve())
        pub = priv.get_public_key()
        mA = b"Super cool message A"
        mB = b"Super cool message B"
        signer = ECDSA(priv, pub)
        k = randbelow(curve.n-1)
        sigA = signer.sign(mA, k)
        # k is shared, thus a private key can be leaked. (Same with regular DSA)
        sigB = signer.sign(mB, k)
        private = ECDSA.find_private_key_when_nonce_is_reused(sigA, mA, sigB, mB, pub)
        self.assertEqual(priv.private_key, private)



from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.LWE import LWEKey, LWE
import secrets
class TestLWE(TestCase):
    def test_encrypt(self):
        key_pair = LWEKey.new(1024)
        m = secrets.token_bytes(4)
        cipher = LWE(key_pair)
        ciphertxt = LWE.encrypt_message(key_pair.public, m)
        plaintext = cipher.decrypt_message(ciphertxt)
        self.assertEqual(plaintext, m)




from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ElGamal import ElGamal
from AsymmetricEncryptions import BytesAndInts, Exceptions
from secrets import token_bytes
class TestElGamal(TestCase):

    def test_generate_key_pair(self):
        priv, pub = ElGamal.generate_key_pair(256)
        self.assertEqual(pub.y, pow(pub.g, priv.x, pub.p))

    def test_generate_key_pair_type(self):
        self.assertRaises(TypeError, ElGamal.generate_key_pair, "test")
        self.assertRaises(TypeError, ElGamal.generate_key_pair, b"test")

    def test_generate_key_pair_values(self):
        self.assertRaises(ValueError, ElGamal.generate_key_pair, -2)
        self.assertRaises(ValueError, ElGamal.generate_key_pair, 5)

    def test_encrypt(self):
        priv, pub = ElGamal.generate_key_pair(256)
        cipher = ElGamal(pub)
        c = cipher.encrypt(b"\x03")
        self.assertEqual(b"\x03", ElGamal(priv).decrypt(c))

    def test_encrypt_type(self):
        priv, pub = ElGamal.generate_key_pair(256)
        cipher = ElGamal(pub)
        self.assertRaises(TypeError, cipher.encrypt, "test")

    def test_load(self):
        priv, pub = ElGamal.generate_key_pair(256)
        cipher = ElGamal(pub)
        c = cipher.encrypt(b"test")
        cipher = ElGamal(priv)
        self.assertEqual(b"test", cipher.decrypt(c))

    def test_sign_verify(self):
        priv, pub = ElGamal.generate_key_pair(1024)
        cipher = ElGamal(priv)
        b = token_bytes(8)
        sig = cipher.sign(b)
        self.assertEqual(cipher.verify(sig), None)
        priv, pub = ElGamal.generate_key_pair(1024)
        cipher = ElGamal(priv)
        b = token_bytes(16)
        sig = cipher.sign(b)
        fake_sig = [b"t", b"es", b"t"]
        self.assertRaises(Exceptions.MACError, cipher.verify, fake_sig)
        changed_sig = (sig[0][:-1], sig[1], sig[2])
        self.assertRaises(Exceptions.MACError, cipher.verify, changed_sig)





from __future__ import annotations
from unittest import TestCase
from AsymmetricEncryptions import EllipticCurveNISTP256, ECDH, ECKey

class TestECDH(TestCase):

    def test_ecdh(self):
        curve = EllipticCurveNISTP256.get_curve()
        keyA = ECKey.new(curve=curve)
        ecdh = ECDH(keyA)
        A = keyA.public_key
        keyB = ECKey.new(curve=curve)
        B = keyB.public_key
        shared_key_alice = ecdh.Stage1(B)
        shared_key_bob = ECDH.Stage2(keyB, A)
        self.assertEqual(shared_key_alice, shared_key_bob)




from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.LWE import LWEKey
from AsymmetricEncryptions import KDF
from unittest import TestCase
import secrets

class TestLWEKey(TestCase):

    def test_export_load(self):
        s_key = KDF.derive_key(secrets.token_bytes(16))
        priv_key = LWEKey.new(128)
        priv_key.export("TestFiles\\LWEKeyTestFileUnprotected.key", b"")
        new_key = LWEKey.load("TestFiles\\LWEKeyTestFileUnprotected.key", b"")
        self.assertEqual(new_key, priv_key)
        priv_key = LWEKey.new(128)
        priv_key.export("TestFiles\\LWEKeyTestFileProtected.key", s_key)
        new_key = LWEKey.load("TestFiles\\LWEKeyTestFileProtected.key", s_key)
        self.assertEqual(new_key, priv_key)





from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.DSA import DSA
from AsymmetricEncryptions.Exceptions import MACError

class TestRSA(TestCase):
    def test_sign_and_verify(self):
        message: bytes = b"DSA test"
        priv, pub = DSA.generate_key_pair(1024)
        cipher = DSA(priv)
        sig = cipher.sign(message)
        cipher = DSA(pub)
        self.assertEqual(cipher.verify(sig, message), None)

    def test_sign_and_verify_err(self):
        message: bytes = b"DSA test"
        priv, pub = DSA.generate_key_pair(1024)
        cipher = DSA(priv)
        sig = cipher.sign(message)
        cipher = DSA(pub)
        self.assertRaises(MACError,cipher.verify,[sig[0] -1, sig[1]], message)




from unittest import TestCase
from AsymmetricEncryptions.General.BytesAndInts import BytesAndInts


class TestBytesAndInts(TestCase):

    # ------------------------- int2byte Start ---------------------
    def test_int2byte(self) -> None:
        # When i >= 0
        self.assertEqual(BytesAndInts.int2Byte(478560413032), b"hello")
        self.assertEqual(BytesAndInts.int2Byte(48), b"0")
        self.assertEqual(BytesAndInts.int2Byte(0), b"")

    def test_int2byte_value(self) -> None:
        # When i < 0
        self.assertRaises(ValueError, BytesAndInts.int2Byte, -1)
        self.assertRaises(ValueError, BytesAndInts.int2Byte, -2)
    # ------------------------- int2byte End ---------------------

    # ------------------------- byte2int Start ---------------------
    def test_byte2int(self):
        # When b is bytes
        self.assertEqual(BytesAndInts.byte2Int(b"hello"), 478560413032)
        self.assertEqual(BytesAndInts.byte2Int(b"0"), 48)
        self.assertEqual(BytesAndInts.byte2Int(b""), 0)
    # ------------------------- byte2int End ---------------------



from AsymmetricEncryptions.Protocols.SchnorrPOK import POK
from AsymmetricEncryptions.PublicPrivateKey import DLIES
from unittest import TestCase

class TestPOK(TestCase):

    def test_proof(self):
        priv, pub = DLIES.generate_key_pair(1024)
        m = b"test"
        prover = POK(priv)
        proof = prover.prove(m)
        self.assertEqual(POK.verify(proof), True)


from __future__ import annotations

class GaloisField:

    def __init__(self, v: [int], primitive_poly: [int], deg: int = 128):
        self.v: [int] = v
        self.deg: int = deg
        self.primitive_poly: [int] = primitive_poly

    def set_v(self, v):
        self.v = v

    @staticmethod
    def pad(a: GaloisField, b: GaloisField) -> (GaloisField, GaloisField):
        if len(a.v) == len(b.v): return a, b
        if len(a.v) < len(b.v):
            # a needs padding
            p: int = len(b.v) - len(a.v)
            v: list[int] = a.v.copy()
            [v.insert(0, 0) for _ in range(p)]
            a.set_v(v)
        if len(a.v) > len(b.v):
            # b needs padding
            p: int = len(a.v) - len(b.v)
            v: list[int] = b.v.copy()
            [v.insert(0, 0) for _ in range(p)]
            b.set_v(v)
        return a, b


    def __add__(self, other: GaloisField) -> GaloisField:
        a: GaloisField = self.copy()
        b: GaloisField = other.copy()
        a, b = GaloisField.pad(a, b)
        return GaloisField([a.v[i] ^ b.v[i] for i in range(len(a.v))], self.primitive_poly, deg=self.deg)

    def strip0(self) -> GaloisField:
        v = self.v.copy()
        while v[0] == 0:
            v.pop(0)
        return GaloisField(v, self.primitive_poly, deg=self.deg)


    def __sub__(self, other: GaloisField) -> GaloisField:
        return self + other

    def copy(self):
        return GaloisField(self.v.copy(), self.primitive_poly.copy(), self.deg)

    def __mul__(self, other: GaloisField) -> GaloisField:
        r = GaloisField([0] * len(self.v), self.primitive_poly, self.deg)
        a = self.copy()
        b = other.copy()
        while 1 in a.v:
            r = r + b
            a = a >> 1
            b = b << 1
            a, b = GaloisField.pad(a, b)
        return r

    def __lshift__(self, other: int) -> GaloisField:
        v = self.v.copy()
        [v.append(0) for _ in range(other)]
        return GaloisField(v, self.primitive_poly, self.deg)

    def __rshift__(self, other: int) -> GaloisField:
        v: list[int] = self.v.copy()
        [v.insert(0, 0) for _ in range(other)]
        v = v[:-other]
        return GaloisField(v, self.primitive_poly, self.deg)

    def __str__(self):
        return str(self.v)



from __future__ import annotations
from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ECC import ECIES, ECKey, ECPoint, EllipticCurveNISTP256
from secrets import token_bytes

class TestECIES(TestCase):


    def test_encrypt(self):
        curve = EllipticCurveNISTP256.get_curve()
        priv = ECKey.new(curve)
        pub = priv.public_key
        m = token_bytes(16)
        e = ECIES.encrypt(m, pub)
        d = ECIES.decrypt(e, priv)
        self.assertEqual(m, d)
    
    def test_export(self):
        curve = EllipticCurveNISTP256.get_curve()
        priv = ECKey.new(curve)
        priv.export("TestFiles\\ECKeyTestFileUnprotected.key", b"")
        loaded_priv = ECKey.load("TestFiles\\ECKeyTestFileUnprotected.key", b"")
        self.assertEqual(priv, loaded_priv)
        priv: ECKey = ECKey.new(curve)
        priv.export("TestFiles\\ECKeyTestFileProtected.key", b"Pass")
        loaded_priv = ECKey.load("TestFiles\\ECKeyTestFileProtected.key", b"Pass")
        self.assertEqual(priv, loaded_priv)



import json
import secrets

from AsymmetricEncryptions.Protocols import KDF
from base64 import b64encode
from datetime import datetime, timedelta

class AS:
    # Authentication server
    def __init__(self, clients: dict[str, bytes], enc_func: callable, KTGS: bytes, ticket_life_time: datetime):
        self.KTGS = KTGS
        self.clients = clients
        self.enc = enc_func
        self.ticket_life_time = ticket_life_time

    def reply(self, approach_request: str):
        try:
            d = json.loads(approach_request)
        except Exception:
            return None

        if not d["client_id"] in self.clients: return None
        clients_key: bytes = KDF.derive_key(self.clients[d["client_id"]])
        Kc_TGS: bytes = KDF.derive_key(secrets.token_bytes(32))
        Kc_TGS = b64encode(Kc_TGS).decode()
        now = datetime.now()
        # year can't be 0 so I do -1
        time_change = timedelta(days=self.ticket_life_time.day + ((self.ticket_life_time.year - 1) * 365), minutes=self.ticket_life_time.minute, hours=self.ticket_life_time.hour)

        life_time = now + time_change
        life_time = life_time.replace(microsecond=0).strftime("%Y-%m-%d %H:%M:%S")
        msgA_unenc = {"nonce": d["nonce"], "lifetime": life_time, "Kc_TGS": Kc_TGS, "service_id": d["service_id"]}
        msgA_json = json.dumps(msgA_unenc)
        msgA: bytes = self.enc(msgA_json, clients_key)
        msgB_unenc = {"client_id": d["client_id"], "Kc_TGS": Kc_TGS, "lifetime": life_time, "service_id": d["service_id"], "nonce": d["nonce"]}
        msgB_json = json.dumps(msgB_unenc)
        TGT: bytes = self.enc(msgB_json, self.KTGS)
        return msgA, TGT






from unittest import TestCase
from secrets import randbelow

class TestMPCAdd(TestCase):

    def test_add(self):
        # Addition of A B and C over modulus n, if n == 0 then there's no modulus
        from AsymmetricEncryptions.MPC.MPC_Addition import MPCAddition
        n = randbelow(40)
        m = randbelow(1024)
        rs = [randbelow(m) for _ in range(n)]
        parties = [MPCAddition(rs[i], n, m) for i in range(n)]
        shares = [parties[i].get_shares() for i in range(len(parties))]
        total = list(zip(*shares))
        ss = [parties[i].s(total[i]) for i in range(n)]
        su = sum(ss) % m
        self.assertEqual(su, sum(rs) % m)

    def test_add_without_mod(self):
        # Addition of A B and C over modulus n, if n == 0 then there's no modulus
        from AsymmetricEncryptions.MPC.MPC_Addition import MPCAddition
        n = randbelow(40)
        m = 0
        rs = [randbelow(1024) for _ in range(n)]
        parties = [MPCAddition(rs[i], n, m) for i in range(n)]
        shares = [parties[i].get_shares() for i in range(len(parties))]
        total = list(zip(*shares))
        ss = [parties[i].s(total[i]) for i in range(n)]
        su = sum(ss)
        self.assertEqual(su, sum(rs))


from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ElGamal import ElGamalKey

class TestElGamalKey(TestCase):
    def test_new(self):
        priv = ElGamalKey.new(256)
        pub: ElGamalKey = priv.public
        self.assertEqual(pub.y, pow(priv.g, priv.x, priv.p))

    def test_new_values(self):
        self.assertRaises(ValueError, ElGamalKey.new, -1)
        self.assertRaises(ValueError, ElGamalKey.new, 4209)

    def test_export_and_load(self):
        priv: ElGamalKey = ElGamalKey.new(256)
        priv.export("TestFiles\\ElGamalKeyTestFileUnprotected.key", b"")
        loaded_priv = ElGamalKey.load("TestFiles\\ElGamalKeyTestFileUnprotected.key", b"")
        self.assertEqual(priv, loaded_priv)
        priv.export("TestFiles\\ElGamalKeyTestFileProtected.key", b"Pass")
        loaded_priv = ElGamalKey.load("TestFiles\\ElGamalKeyTestFileProtected.key", b"Pass")
        self.assertEqual(priv, loaded_priv)






from __future__ import annotations
from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.ECC import EllipticCurveNISTP256, ECPoint

class TestECCurve(TestCase):

    def test_g(self):
        self.assertEqual(EllipticCurveNISTP256.get_curve().g(), ECPoint(EllipticCurveNISTP256.get_curve(), EllipticCurveNISTP256.g_x, EllipticCurveNISTP256.g_y))

    def test_f(self):
        curve = EllipticCurveNISTP256.get_curve()
        self.assertEqual(curve.f(3), 41058363725152142129326129780047268409114441015993725554835256314039467401309)

    def test_f_value(self):
        curve = EllipticCurveNISTP256.get_curve()
        self.assertRaises(ValueError, curve.f, -1)




__all__ = ["Exceptions", "NeededValueIsNull", "UnsafeEncryptionFunction", "MACError", "DeprecatedClass", "ReplayAttack"]
from .Exceptions import NeededValueIsNull, UnsafeEncryptionFunction, MACError, DeprecatedClass, ReplayAttack



from AsymmetricEncryptions.PublicPrivateKey.RSA import RSAKey, RSA
from AsymmetricEncryptions.Protocols.RingSignatures import RingSignatures
import secrets
from unittest import TestCase

class TestRingSignature(TestCase):
    def test_sign_and_verify(self):
        # Gen
        ksize = 1024
        gsize = secrets.randbelow(10)
        while gsize < 3:
            gsize = secrets.randbelow(10)
        og_priv, _ = RSA.generate_key_pair(ksize)
        og_priv: RSAKey
        keys = []
        for i in range(gsize - 1):
            _, pub = RSA.generate_key_pair(ksize)
            keys.append(pub)
        msg1 = secrets.token_bytes(16)
        p = secrets.randbelow(len(keys))
        keys.insert(p, og_priv)
        r = RingSignatures(keys)
        sig, RK = r.sign_message(msg1, p)
        ver = RingSignatures(RK).verify_message(msg1, sig)
        self.assertEqual(ver, True)


from __future__ import annotations
from unittest import TestCase
from AsymmetricEncryptions import ECPoint, EllipticCurveNISTP256

class TestECPoint(TestCase):

    def test_copy(self):
        curve = EllipticCurveNISTP256.get_curve()
        p1: ECPoint = curve.g() * 2
        p1_copy = p1.copy()
        self.assertEqual(p1_copy, p1)

    def test_add(self):
        curve = EllipticCurveNISTP256.get_curve()
        p1: ECPoint = curve.g() * 2
        p2: ECPoint = curve.g() + curve.g()
        self.assertEqual(p1, p2)





from unittest import TestCase
from AsymmetricEncryptions.General.XOR import XOR

class TestXOR(TestCase):

    def test_repeated_key_xor(self):
        self.assertEqual(XOR.repeated_key_xor(b"1", b"1"), b"\x00")
        self.assertEqual(XOR.repeated_key_xor(b"2", b"1"), b'\x03')
        self.assertEqual(XOR.repeated_key_xor(b"\x00", b"1"), b'1')
        self.assertEqual(XOR.repeated_key_xor(XOR.repeated_key_xor(b"msg", b"key"), b"key"), b'msg')

    def test_repeated_key_xor_with_scrypt_kdf(self):
        self.assertEqual(XOR.repeated_key_xor_with_scrypt_kdf(XOR.repeated_key_xor_with_scrypt_kdf(b"msg", b"key"), b"key"), b'msg')








class NeededValueIsNull(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)

class UnsafeEncryptionFunction(Exception):
    def __init__(self, func_name: str, msg: str):
        super().__init__(f"The function {func_name} is an unsafe function, error message:{msg}")

class ReplayAttack(Exception):
    def __init__(self):
        super().__init__(f"Replay Attack!")

class MACError(Exception):

    def __init__(self, msg: str, mac_type: str = "HMAC"):
        super().__init__(f"A MAC of type {mac_type} is not matching. Error message: {msg}")

class DeprecatedClass(Exception):

    def __init__(self):
        super().__init__(f"The Function or Class is deprecated and is to not be used until fixed!")

    @staticmethod
    def DeprecatedFunc(*args, **kwargs) -> callable:
        def wrapper(*args, **kwargs) -> callable:
            raise DeprecatedClass()
        return wrapper



from AsymmetricEncryptions.PublicPrivateKey.ECC import ECKey, ECPoint, ECCurve
from AsymmetricEncryptions.General import BytesAndInts
from hashlib import sha256
from warnings import warn
import secrets

class ECDSA:

    def __init__(self, private_key: ECKey, public_key: ECKey, nonces: set = None):
        self.private_key = private_key
        self.public_key = public_key
        self.nonces = set()
        if nonces is not None:
            self.nonces = nonces

    def sign(self, m: bytes, k=None):
        e: int = BytesAndInts.byte2Int(sha256(m).digest())
        if k is None:
            k: int = secrets.randbelow(self.private_key.curve.n - 1)
            hash_k = sha256(f"{k}".encode()).digest()
            if hash_k in self.nonces: return self.sign(m)
            self.nonces.add(hash_k)
        else: warn(f"\nIf k is static and not random the security is compromised.\n\nm: {m}, k: {k}\n\n")

        r: ECPoint = self.private_key.curve.g() * k
        r: int = r.x
        if r == 0:
            return self.sign(m)
        inv_k: int = self.private_key.curve.find_inverse(k)
        s = (inv_k * (e + r*self.private_key.private_key)) % self.private_key.curve.n
        if s == 0: return self.sign(m)
        return r, s

    @staticmethod
    def verify(m: bytes, public_key: ECKey, signature: tuple[int, int]) -> bool:
        if not  public_key.public_key != public_key.curve.infinity(): return False
        if not  public_key.curve.is_on_curve(public_key.public_key): return False
        if not public_key.public_key * public_key.curve.n == public_key.curve.infinity(): return False
        r, s = signature
        if not 1 < r < (public_key.curve.n - 1) or not 1 < s < (public_key.curve.n - 1): return False
        e: int = BytesAndInts.byte2Int(sha256(m).digest())
        inv_s = public_key.curve.find_inverse(s)
        u1: int = (e * inv_s) % public_key.curve.n
        u2: int = (r * inv_s) % public_key.curve.n
        G: ECPoint = public_key.curve.g()
        w = G * u1 + public_key.public_key * u2
        if w == public_key.curve.infinity(): return False
        return r == w.x

    @staticmethod
    def find_private_key_when_nonce_is_reused(sigA: tuple[int, int], mA: bytes, sigB: tuple[int, int], mB: bytes, public_key: ECKey):
        rA, sA = sigA
        rB, sB = sigB
        assert rA == rB and sA != sB
        eA: int = BytesAndInts.byte2Int(sha256(mA).digest())
        eB: int = BytesAndInts.byte2Int(sha256(mB).digest())
        n: int = public_key.curve.n
        k = ((eA - eB) % n) * pow((sA - sB) % n, -1, n)
        priv = ((sA * k - eA) % n) * pow(rA, -1, n)
        priv %= n
        return priv



from __future__ import annotations
from AsymmetricEncryptions.General.XOR import XOR
from abc import ABC, abstractmethod

class IKey(ABC):

    @staticmethod
    @abstractmethod
    def new(nBit: int):
        raise NotImplementedError("Function should be implemented inside of class")

    @abstractmethod
    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        raise NotImplementedError("Function should be implemented inside of class")

    @staticmethod
    @abstractmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> IKey:
        raise NotImplementedError("Function should be implemented inside of class")





from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSA
from AsymmetricEncryptions import BytesAndInts, Exceptions
from secrets import token_bytes
class TestRSA(TestCase):

    def test_generate_key_pair(self):
        priv, pub = RSA.generate_key_pair(256)
        self.assertEqual((pub.e * priv.d) % priv.tot_n, 1)

    def test_generate_key_pair_type(self):
        self.assertRaises(TypeError, RSA.generate_key_pair, "test")
        self.assertRaises(TypeError, RSA.generate_key_pair, b"test")

    def test_generate_key_pair_values(self):
        self.assertRaises(ValueError, RSA.generate_key_pair, -2)
        self.assertRaises(ValueError, RSA.generate_key_pair, 5)

    def test_encrypt(self):
        priv, pub = RSA.generate_key_pair(256)
        cipher = RSA(pub)
        c = cipher.encrypt(b"\x03")
        self.assertEqual(c, BytesAndInts.int2Byte(pow(3, pub.e, pub.n)))

    def test_encrypt_type(self):
        priv, pub = RSA.generate_key_pair(256)
        cipher = RSA(pub)
        self.assertRaises(TypeError, cipher.encrypt, "test")

    def test_decrypt(self):
        priv, pub = RSA.generate_key_pair(256)
        cipher = RSA(pub)
        c = cipher.encrypt(b"test")
        cipher = RSA(priv)
        self.assertEqual(b"test", cipher.decrypt(c))

    def test_sign_verify(self):
        priv, pub = RSA.generate_key_pair(256)
        cipher = RSA(priv)
        b = token_bytes(16)
        sig = cipher.sign(b)
        self.assertEqual(cipher.verify(sig, b), None)
        priv, pub = RSA.generate_key_pair(256)
        cipher = RSA(priv)
        b = token_bytes(16)
        sig = cipher.sign(b)
        self.assertRaises(Exceptions.MACError, cipher.verify,sig[:-1], b)




import json
from base64 import b64decode
from datetime import datetime

from .KDC import KDC
from AsymmetricEncryptions.General.XOR import XOR
from AsymmetricEncryptions.General.Feistel_sha256 import FeistelSha256
from AsymmetricEncryptions.Protocols.KDF import KDF

class KerberosService:
    def __init__(self, Ks: bytes, symmetric_enc_func=FeistelSha256.get_feistel().encrypt, symmetric_dec_func=FeistelSha256.get_feistel().decrypt):
        self.enc = KDC.add_hmac(symmetric_enc_func)
        self.dec = KDC.dec_with_hmac(symmetric_dec_func)
        self.ks = KDF.derive_key(Ks)
        self.clients = {}

    def confirm(self, ticket: bytes, msgG: bytes):
        msgE = self.dec(ticket, self.ks)
        ticket = json.loads(msgE)
        Kc_s = b64decode(ticket["Kc_s"].encode())
        client_id = ticket["client_id"]
        if client_id in self.clients: return None
        msgG = self.dec(msgG, Kc_s).decode()
        msgG = json.loads(msgG)
        if msgG["client_id"] != client_id: return None
        self.clients[client_id] = Kc_s
        msgH = json.dumps({"timestamp": datetime.now().replace(microsecond=0).strftime("%Y-%m-%d %H:%M:%S")}).encode()
        return self.enc(msgH, Kc_s)



from __future__ import division
from __future__ import print_function
import secrets
import functools
from AsymmetricEncryptions.General.PrimeNumberGen import PrimeNumberGen
from hashlib import sha256


class SSS:
    """Code is edited from Wikipedia"""

    def __init__(self, prime: int = pow(2, 10141) - 1, mini = 2, maxi = 3):
        if prime < 2: raise ValueError("Prime must be a positive prime number")
        if prime != pow(2, 10141) - 1:
            if not PrimeNumberGen.isMillerRabinPassed(prime): raise ValueError("Prime must be a positive prime number")
        self._PRIME = prime
        self._RINT: int = functools.partial(secrets.SystemRandom().randint, 0)
        self.mini = mini
        self.maxi = maxi

    @staticmethod
    def _eval_at(poly, x, prime):
        """Evaluates polynomial (coefficient tuple) at x, used to generate a
        shamir pool in make_random_shares below.
        """
        accum = 0
        for coeff in reversed(poly):
            accum *= x
            accum += coeff
            accum %= prime
        return accum

    def make_random_shares(self, secret: bytes, minimum: int, shares: int) -> list[tuple[int, int]]:
        from AsymmetricEncryptions import BytesAndInts
        """
        Generates a random shamir pool for a given secret, returns share points.
        """
        if minimum < 2 or shares < 3: raise ValueError("Minimum and Shares must be larger than 2, 3")
        secret = BytesAndInts.byte2Int(secret)
        prime = self._PRIME
        if secret > prime:
            raise ValueError("The secret is too large!")
        if minimum > shares:
            raise ValueError("Pool secret would be irrecoverable.")
        self.mini = minimum
        self.maxi = shares
        poly = [secret] + [self._RINT(prime - 1) for _ in range(minimum - 1)]
        points = [(i, SSS._eval_at(poly, i, prime))
                  for i in range(1, shares + 1)]
        return points

    def get_prime(self) -> int:
        return self._PRIME


    def convert_list_of_bytes_and_points_to_points(self, l: list[tuple[int, int] or bytes]) -> list[tuple[int, int]]:
        r = []
        for obj in l:
            if isinstance(obj, bytes):
                r.append(SSS.convert_bytes_to_point(obj, self._PRIME))
            else:
                r.append(obj)
        return r

    def generate_for_public(self, secret: bytes, custom: list[tuple[int, int] or bytes]) -> tuple[int, int]:
        from AsymmetricEncryptions import BytesAndInts
        secret = BytesAndInts.byte2Int(secret)
        prime = self._PRIME
        if secret > prime:
            raise ValueError("The secret is too large!")
        if len(custom) == 0: raise ValueError("The secret would be irrecoverable")
        r = self._RINT(prime - 1)
        custom = self.convert_list_of_bytes_and_points_to_points(custom)
        custom.append((0, secret))
        x_s, y_s = zip(*custom)
        point = (r, self._lagrange_interpolate(r, x_s, y_s, prime))
        return point



    @staticmethod
    def convert_bytes_to_point(b: bytes, prime: int) -> tuple[int, int]:
        from AsymmetricEncryptions import BytesAndInts
        return BytesAndInts.byte2Int(b) % prime, BytesAndInts.byte2Int(sha256(b).digest()) % prime



    @staticmethod
    def _extended_gcd(a, b):
        """
        Division in integers modulus p means finding the inverse of the
        denominator modulo p and then multiplying the numerator by this
        inverse (Note: inverse of A is B such that A*B % p == 1). This can
        be computed via the extended Euclidean algorithm
        http://en.wikipedia.org/wiki/Modular_multiplicative_inverse#Computation
        """
        x = 0
        last_x = 1
        y = 1
        last_y = 0
        while b != 0:
            quot = a // b
            a, b = b, a % b
            x, last_x = last_x - quot * x, x
            y, last_y = last_y - quot * y, y
        return last_x, last_y

    @staticmethod
    def _divmod(num, den, p):
        """Compute num / den modulo prime p

        To explain this, the result will be such that:
        den * _divmod(num, den, p) % p == num
        """
        inv, _ = SSS._extended_gcd(den, p)
        return num * inv

    def _lagrange_interpolate(self, x, x_s, y_s, p):
        """
        Find the y-value for the given x, given n (x, y) points;
        k points will define a polynomial of up to kth order.
        """
        k = len(x_s)
        assert k == len(set(x_s)), "points must be distinct"

        def PI(vals):  # upper-case PI -- product of inputs
            accum = 1
            for v in vals:
                accum *= v
            return accum
        nums = []  # avoid inexact division
        dens = []
        for i in range(k):
            others = list(x_s)
            cur = others.pop(i)
            nums.append(PI(x - o for o in others))
            dens.append(PI(cur - o for o in others))
        den = PI(dens)
        num = sum([SSS._divmod(nums[i] * den * y_s[i] % p, dens[i], p)
                   for i in range(k)])
        return (SSS._divmod(num, den, p) + p) % p


    def recover_secret(self, shares: list[tuple[int, int]]) -> bytes:
        """
        Recover the secret from share points
        (points (x,y) on the polynomial).
        """
        from AsymmetricEncryptions import BytesAndInts
        shares = self.convert_list_of_bytes_and_points_to_points(shares)
        prime = self._PRIME
        if len(shares) < self.mini:
            raise ValueError("need at least t shares")
        x_s, y_s = zip(*shares)
        return BytesAndInts.int2Byte(self._lagrange_interpolate(0, x_s, y_s, prime))


from unittest import TestCase
# This isn't really secure since I didn't implement the nonce thing fully as well as the ticket lifetime, but this is a showcase
from AsymmetricEncryptions.Protocols.Kerberos import KDC, KerberosService, KerberosClient
# Make sure the symmetric encryption function is AES (Make a wrapper function, use other library like PyCryptodom or Cryptography)

class TestKerberos(TestCase):
    def test_kerberos(self):
        clients = {"Alice": b"Alice", "Bob": b"Bob"}
        services = {"S": b"Secret", "CRM": b"Super secret passwOrd"}

        serviceS = KerberosService(b"Secret")
        CRM = KerberosService(b"Super secret passwOrd")

        kdc = KDC(clients, services, b"secret password")
        clientAlice = KerberosClient("Alice", b"Alice")

        approach = clientAlice.approach_AS("S")

        msgA, TGT = kdc.AS_response(approach[0])

        approach_TGS, Kc_tgs = clientAlice.approach_TGS(msgA, TGT)

        ticket, msgF = kdc.TGS_response(*approach_TGS)

        approach_service, Kc_s = clientAlice.approach_service(ticket, msgF, Kc_tgs)

        serviceS.confirm(*approach_service)

        clientBob = KerberosClient("Bob", b"Bob")

        approach = clientBob.approach_AS("CRM")

        msgA, TGT = kdc.AS_response(approach[0])

        approach_TGS, Kc_tgs = clientBob.approach_TGS(msgA, TGT)

        ticket, msgF = kdc.TGS_response(*approach_TGS)

        approach_service, Kc_s = clientBob.approach_service(ticket, msgF, Kc_tgs)

        CRM.confirm(*approach_service)
        self.assertEqual(None, None)
        try:
            clientCarol = KerberosClient("Bob", b"Carol")

            approach = clientCarol.approach_AS("CRM")

            msgA, TGT = kdc.AS_response(approach[0])

            approach_TGS, Kc_tgs = clientCarol.approach_TGS(msgA, TGT)

            ticket, msgF = kdc.TGS_response(*approach_TGS)

            approach_service, Kc_s = clientCarol.approach_service(ticket, msgF, Kc_tgs)
        except Exception:
            self.assertEqual(None, None)
        else:
            self.assertEqual(True, False)



from unittest import TestCase
from AsymmetricEncryptions.PublicPrivateKey.DLIES import DLIESKey

class TestDLIESKey(TestCase):
    def test_new(self):
        priv = DLIESKey.new(256)
        pub: DLIESKey = priv.public
        self.assertEqual(pub.y, pow(priv.g, priv.x, priv.n))

    def test_new_values(self):
        self.assertRaises(ValueError, DLIESKey.new, -1)
        self.assertRaises(ValueError, DLIESKey.new, 4209)

    def test_export_and_load(self):
        priv: DLIESKey = DLIESKey.new(256)
        priv.export("TestFiles\\DLIESKeyTestFileUnprotected.key", b"")
        loaded_priv = DLIESKey.load("TestFiles\\DLIESKeyTestFileUnprotected.key", b"")
        self.assertEqual(priv, loaded_priv)
        priv: DLIESKey = DLIESKey.new(256)
        priv.export("TestFiles\\DLIESKeyTestFileProtected.key", b"Pass")
        loaded_priv = DLIESKey.load("TestFiles\\DLIESKeyTestFileProtected.key", b"Pass")
        self.assertEqual(priv, loaded_priv)






from AsymmetricEncryptions.Protocols import KDF, PKCS7
from AsymmetricEncryptions.General import XOR
from AsymmetricEncryptions.General.Feistel_sha256 import FeistelSha256
from AsymmetricEncryptions.Exceptions import MACError
import json
from secrets import compare_digest, token_bytes
import hmac
import base64
from hashlib import sha256

class Exportation:
    """
    Handles the exportation of keys. Nothing in here is done by regulations though.
    """

    @staticmethod
    def sha_wrapper( b: bytes) -> bytes:
        return sha256(b).digest()

    @staticmethod
    def export(file_name: str, data_dict: dict, pwd: bytes, *, exportation_func=FeistelSha256.get_feistel().encrypt, block_size: int = 32, isTesting: bool = False) -> bytes:
        """
        Encrypts and exports data that can be converted to json.
        @param file_name: The file that will be exported into
        @param data_dict: The data that will be exported
        @param pwd: The passphrase that will be used in the KDF and then encryption part of the export
        @param exportation_func: The encryption function used (I'm using XOR or OTP)
        @return: what's in the file
        @param block_size: The symmetric encryption block size
        @param isTesting if the test is active it won't save
        """
        key: bytes = KDF.derive_key(pwd)
        jData: bytes = json.dumps(data_dict).encode("utf-8")
        randomness: bytes = Exportation.sha_wrapper(token_bytes(32))
        xored: bytes = XOR.repeated_key_xor(jData, randomness)
        xored: bytes = PKCS7(block_size).pad(xored)
        write_data: bytes = exportation_func(xored, key)
        mac: hmac = hmac.new(key=key, msg=jData, digestmod="sha512")
        final_data: bytes = mac.digest() + write_data + randomness
        final_data = base64.b64encode(final_data)
        if not isTesting:
            with open(file_name, "wb") as f:
                f.write(final_data)
        return final_data

    @staticmethod
    def load(file_name: str, pwd: bytes, *, dec_func=FeistelSha256.get_feistel().decrypt, block_size: int = 32, isTesting: bool = False, dataIfTesting: bytes = b"") -> dict:
        """
        Loads data from an encrypted file
        @param file_name: The encrypted file name
        @param pwd: The passphrase used as a key
        @param dec_func: The opposite of the encryption function
        @return: The data as a dict
        @param block_size: The symmetric decryption block size
        @param isTesting: if testing then the dataIfTesting will take over instead of reading the file
        @param dataIfTesting: if testing then the dataIfTesting will take over instead of reading the file
        """
        key: bytes = KDF.derive_key(pwd)
        final_data: bytes = dataIfTesting
        if not isTesting:
            with open(file_name, "rb") as f:
                final_data: bytes = f.read()
        final_data = base64.b64decode(final_data)
        dMac: bytes = final_data[:64]
        read_data: bytes = final_data[64:-32]
        randomness: bytes = final_data[-32:]
        jData: bytes = dec_func(read_data, key)
        jData: bytes = PKCS7(block_size).unpad(jData)
        jData = XOR.repeated_key_xor(jData, randomness)
        mac: hmac = hmac.new(key=key, msg=jData, digestmod="sha512")
        if not compare_digest(mac.digest(), dMac):
            raise MACError("The macs don't match!")
        return json.loads(jData.decode())
